"""
Provides functionality for lazily deserializing classes field-by-field.

When deserializing an object, rather than deserializing all of its fields
immediately, some additional fields/methods are added:

- `_uniserde_field_definitions_`: A dictionary mapping field names to tuples of
  (field document name, field type).

- `_uniserde_remaining_fields_`: A dictionary mapping field document names to
  unparsed field values.

- `_uniserde_serdeserializer_`: The serdeserializer instance that was used to
  deserialize the object.

- `__getattr__`: A function that deserializes fields when they are first
  accessed, and then caches them in the instance.

Note that this approach allows for doing a lot of gruntwork up-front: The class,
along with all of its fields, and the `__getattr__` method only have to be
created once and can be reused for all future deserializations of that class.
"""

from __future__ import annotations

import inspect
from typing import *  # type: ignore

from . import caching_serdeserializer
from .common import SerdeError

_FIELD_MAP_CACHE: dict[
    tuple[Type, Type[caching_serdeserializer.CachingSerDeserializer]], Type
] = {}


def _lazy_getattr(self, name: str) -> Any:
    # Fetch the field definitions. This will fail if this particular instance
    # isn't lazy
    try:
        field_definitions = vars(self)["_uniserde_field_definitions_"]
    except KeyError:
        raise AttributeError(name) from None

    # See if there is a field definition for this field. This will fail if the
    # field doesn't exist
    try:
        field_doc_name, field_type = field_definitions[name]
    except KeyError:
        raise AttributeError(name) from None

    # Get the field value
    try:
        field_raw_value = self._uniserde_remaining_fields_.pop(field_doc_name)
    except KeyError:
        raise SerdeError(f"Missing field `{field_doc_name!r}`") from None

    # Deserialize it
    field_handler = self._uniserde_serdeserializer_._get_handler(field_type)
    parsed_value = field_handler(
        self._uniserde_serdeserializer_,
        field_raw_value,
        field_type,
    )

    # Cache it
    vars(self)[name] = parsed_value

    # Return it
    return parsed_value


def _get_field_map(
    value_type: Type,
    serdeserializer: caching_serdeserializer.CachingSerDeserializer,
) -> dict[str, tuple[str, Type]]:
    # Already cached?
    try:
        return _FIELD_MAP_CACHE[(value_type, type(serdeserializer))]
    except KeyError:
        pass

    # Fetch the fields
    result = {
        field_py_name: (
            field_doc_name,
            field_type,
        )
        for (
            field_py_name,
            field_doc_name,
            field_type,
        ) in serdeserializer._get_class_fields(value_type)
    }

    # Cache the result
    _FIELD_MAP_CACHE[(value_type, type(serdeserializer))] = result

    # Return it
    return result


def can_create_lazy_instance(value_type: Type) -> bool:
    """
    Verify that some conditions are met for creating a lazy instance of the
    given type.
    """

    # The class mustn't have a `__getattr__` method, since that would be
    # overwritten. If it is already overwritten that's obviously also fine.
    try:
        return value_type.__getattr__ is _lazy_getattr
    except AttributeError:
        return True


def create_lazy_instance(
    serialized: dict[str, Any],
    serdeserializer: caching_serdeserializer.CachingSerDeserializer,
    value_type: Type,
) -> Any:
    assert isinstance(serialized, dict), serialized
    assert inspect.isclass(value_type), value_type
    assert serdeserializer._lazy, serdeserializer
    assert can_create_lazy_instance(value_type), value_type

    # Instantiate the result, skipping the constructor
    result = object.__new__(value_type)

    # Set additional, internal fields
    type(result).__getattr__ = _lazy_getattr
    result._uniserde_field_definitions_ = _get_field_map(value_type, serdeserializer)
    result._uniserde_serdeserializer_ = serdeserializer
    result._uniserde_remaining_fields_ = serialized

    # Done
    return result



import enum
import inspect
import types
from abc import ABC, abstractmethod
from typing import *  # type: ignore

from typing_extensions import TypeAlias

from . import case_convert, common, lazy_wrapper, serde_class
from .common import SerdeError

IN = TypeVar("IN")
OUT = TypeVar("OUT")

Handler: TypeAlias = Callable[["CachingSerDeserializer[IN, OUT]", IN, Type[OUT]], OUT]


class CachingSerDeserializer(ABC, Generic[IN, OUT]):
    """
    Serializes/Deserializes objects, while caching any type-specific handlers it
    creates. This provides massive speedups over constantly trying to figure out
    what to do. (~3x compared to the old approach.)

    Despite the name, currently only handles deserialization.
    """

    # Types which are passed through without any processing. Must be defined by
    # subclasses.
    _passthrough_types: Set[Type[IN]]

    # Maps (lazy, type) to handlers. This acts as cache for previously seen
    # types.
    #
    # Must be defined by subclasses. (An empty dict is fine.)
    _handler_cache: dict[tuple[bool, Type[IN]], Handler]

    # If a method of this name is present in a class (and it isn't the one
    # inherited from `Serde`) this will be used for handling that class, rather
    # than the default behavior.
    _override_method_name: str

    def __init_subclass__(cls) -> None:
        # Add handlers for all of the class's passthrough types
        def make_handler_from_passthrough_type(
            passthrough_type: Type,
        ) -> Handler:
            def result(self, value: IN, as_type: Type[OUT]) -> OUT:
                if not isinstance(value, passthrough_type) and not (
                    isinstance(value, int) and passthrough_type is float
                ):
                    raise SerdeError(f"Expected `{passthrough_type}`, got `{value}`")

                return value  # type: ignore

            return result

        for typ in cls._passthrough_types:
            handler = make_handler_from_passthrough_type(typ)
            cls._handler_cache[(True, typ)] = handler
            cls._handler_cache[(False, typ)] = handler

    def __init__(
        self,
        *,
        custom_deserializers: dict[Type, Callable[[IN, Type[OUT]], OUT]],
        lazy: bool,
    ):
        self._custom_deserializers = custom_deserializers
        self._lazy = lazy

    @abstractmethod
    def _get_class_fields(self, value_type: Type) -> Iterable[tuple[str, str, Type]]:
        """
        Return a tuple (python_name, json_name, type) for each field in the
        class.
        """
        raise NotImplementedError()

    def process(self, value: IN, as_type: Type[OUT]) -> OUT:
        # Special case: rapidly handle simple passthrough types to increase
        # performance
        if as_type in self._passthrough_types:
            if not isinstance(value, as_type) and not (
                isinstance(value, int) and as_type is float
            ):
                raise SerdeError(f"Expected `{as_type}`, got `{value}`")

            return value  # type: ignore

        # Otherwise get a handler and use it
        handler = self._get_handler(as_type)
        return handler(self, value, as_type)

    def _get_handler(
        self,
        value_type: Type,
    ) -> Handler:
        # Find a matching serializer
        type_key = common.get_type_key(value_type)

        # Custom handlers take precedence
        try:
            custom_handler = self._custom_deserializers[type_key]
        except KeyError:
            pass
        else:
            return lambda self, value, as_type: custom_handler(value, as_type)

        # Use a cached handler if possible
        try:
            return self._handler_cache[(self._lazy, type_key)]
        except KeyError:
            pass

        # Otherwise create the appropriate handler and cache it for next time
        assert inspect.isclass(value_type), value_type
        handler = self._create_class_handler(value_type)
        self._handler_cache[(self._lazy, type_key)] = handler

        return handler

    def _create_class_handler(
        self,
        value_type: Type,
    ) -> Handler:
        # Case: The class has a custom method for handling it
        #
        # This needs care, depending on whether the method was just overwritten
        # as a regular method, or as a static/class method.
        try:
            override_method = getattr(value_type, self._override_method_name)
        except AttributeError:
            pass
        else:
            serde_class_method = getattr(serde_class.Serde, self._override_method_name)

            try:
                override_method_func = override_method.__func__
            except AttributeError:
                override_method_func = override_method

            if override_method_func is not serde_class_method.__func__:
                return lambda self, value, _type: override_method(value, {})

        # Case enum.Flag
        if issubclass(value_type, enum.Flag):

            def handle_flag(self, value, _type):
                if not isinstance(value, list):
                    raise SerdeError(f"Expected a list, got `{value}`")

                result = value_type(0)

                for item in value:
                    if not isinstance(item, str):
                        raise SerdeError(f"Expected enumeration string, got `{item}`")

                    try:
                        py_name = case_convert.camel_case_to_all_upper(
                            item
                        )  # ValueError if not camel case
                        result |= value_type[py_name]  # ValueError if not in enum
                    except KeyError:
                        raise SerdeError(
                            f"Invalid enumeration value `{item}`"
                        ) from None

                return result

            return handle_flag

        # Case: enum.Enum
        if issubclass(value_type, enum.Enum):

            def handle_enum(self, value, _type):
                if not isinstance(value, str):
                    raise SerdeError(f"Expected enumeration string, got `{value}`")

                try:
                    py_name = case_convert.camel_case_to_all_upper(
                        value
                    )  # ValueError if not camel case
                    return value_type[py_name]  # ValueError if not in enum
                except KeyError:
                    raise SerdeError(f"Invalid enumeration value `{value}`") from None

            return handle_enum

        # Case: Base which is serialized `@as_child`
        if common.should_serialize_as_child(value_type):
            # Precompute a set of all possible classes
            child_classes_and_handlers_by_doc_name = {
                case_convert.upper_camel_case_to_camel_case(cls.__name__): (
                    cls,
                    self._create_fieldwise_class_handler(cls),
                )
                for cls in common.all_subclasses(value_type, True)
            }

            def handle_as_child(self, value, _type):
                # Look up the real type
                try:
                    type_tag = value.pop("type")
                except KeyError:
                    raise SerdeError(f"Object is missing the `type` field") from None

                # Get the class
                try:
                    (
                        child_class,
                        child_class_handler,
                    ) = child_classes_and_handlers_by_doc_name[type_tag]
                except KeyError:
                    raise SerdeError(
                        f"Encountered invalid type tag `{type_tag}`"
                    ) from None

                # Delegate to that class's handler
                return child_class_handler(self, value, child_class)

            # TODO: The generated handler works for all subclasses, but will
            #       only be cached for the one class that has just been
            #       requested. Consider explicitly caching it for all classes.

            return handle_as_child

        # Case: Regular old class
        return self._create_fieldwise_class_handler(value_type)

    def _create_fieldwise_class_handler(self, value_type: Type) -> Handler:
        # Lazy?
        if self._lazy and lazy_wrapper.can_create_lazy_instance(value_type):

            def _handler(self, value, _type):
                return lazy_wrapper.create_lazy_instance(
                    value,
                    self,
                    value_type,
                )

            return _handler

        # Eager
        handler = FieldwiseClassHandler()

        for py_name, doc_name, field_type in self._get_class_fields(value_type):
            handler.add_field(
                py_name,
                doc_name,
                field_type,
                self._get_handler(field_type),
            )

        return handler


class FieldwiseClassHandler:
    """
    Deserializes a class, field by field. The fields and their handlers are
    already passed in the constructor to avoid having to recompute them for each
    instance.
    """

    fields: list[tuple[str, str, Type, Handler]]

    def __init__(self):
        self.fields = []

    def add_field(
        self,
        python_name: str,
        doc_name: str,
        field_type: Type,
        handler: Handler,
    ):
        self.fields.append((python_name, doc_name, field_type, handler))

    def __call__(
        self,
        calling_ser_deserializer: CachingSerDeserializer,
        raw: Any,
        value_type: Type,
    ) -> Any:
        # Make sure the raw value is a dict
        if not isinstance(raw, dict):
            raise SerdeError(f"Expected object, got `{raw!r}`")

        # Create an instance of the class
        result = object.__new__(value_type)
        result_dict = vars(result)

        # Handle all fields
        for py_name, doc_name, field_type, handler in self.fields:
            # Get the raw value
            try:
                raw_value = raw.pop(doc_name)
            except KeyError:
                raise SerdeError(f"Missing field `{doc_name!r}`") from None

            # Process it
            processed_value = handler(calling_ser_deserializer, raw_value, field_type)

            # Store it
            result_dict[py_name] = processed_value

        # Make sure there are no stray fields
        if len(raw) > 0:
            raise SerdeError(
                f"Superfluous object fields `{'`, `'.join(map(str, raw.keys()))}`"
            )

        return result



import dataclasses
import inspect
import types
from typing import *  # type: ignore

T = TypeVar("T")


Recur = Callable[[Any, Type], Any]
Serializer = Callable[[Any, Type, Recur], Any]
Deserializer = Callable[[Any, Type, Recur], Any]


class SerdeError(Exception):
    """
    Signals an error during serialization or deserialization.
    """

    def __init__(self, user_message: str):
        self.user_message = user_message


def as_child(cls: Type[T]) -> Type[T]:
    """
    Marks the class to be serialized as one of its children. This will add an
    additional "type" field in the result, so the child can be deserialized
    properly.

    This decorator applies to children of the class as well, i.e. they will also
    be serialized with the "type" field.
    """
    assert inspect.isclass(cls), cls
    cls.__serde_serialize_as_child__ = cls  # type: ignore
    return cls


def should_serialize_as_child(cls: Type) -> bool:
    """
    Checks whether the given class should be serialized as a child, i.e. it, or
    any parent has been marked with the `as_child` decorator.
    """
    assert inspect.isclass(cls), cls
    return hasattr(cls, "__serde_serialize_as_child__")


def get_type_key(cls: Type) -> Type:
    """
    Given a type, return a more standardized type, suitable for use as a key to
    find serializers/deserializers.
    """

    # See what `get_origin` can do
    result: Any = get_origin(cls)

    if result is None:
        result = cls

    # Convert new-style unions to old-style
    if result is types.UnionType:
        result = Union

    # Pass through the rest
    return result


def common_serialize(
    value: Any,
    value_type: Optional[Type],
    class_serializer: Serializer,
    serializers: dict[Type, Serializer],
    user_serializers: dict[Type, Callable[[Any], Any]],
) -> Any:
    # Find the type
    if value_type is None:
        value_type = type(value)

    # Is there a custom serializer for this class?
    try:
        serializer = user_serializers[value_type]
    except KeyError:
        pass
    else:
        return serializer(value)

    # Find a matching serializer
    key = get_type_key(value_type)

    try:
        serializer = serializers[key]
    except KeyError:
        if inspect.isclass(value_type):
            serializer = class_serializer
        else:
            raise ValueError(f"Unsupported field of type {value_type}") from None

    # Define the recursion function
    def recur(value: Any, value_type: Type) -> Any:
        return common_serialize(
            value,
            value_type,
            class_serializer,
            serializers,
            user_serializers,
        )

    # Apply it
    return serializer(value, value_type, recur)


def common_deserialize(
    value: Any,
    value_type: Type,
    class_deserializer: Serializer,
    deserializers: dict[Type, Deserializer],
    user_deserializers: dict[Type, Callable[[Any], Any]],
) -> Any:
    # Is there a custom deserializer for this class?
    try:
        deserializer = user_deserializers[value_type]
    except KeyError:
        pass
    else:
        return deserializer(value)

    # Find a matching serializer
    key = get_type_key(value_type)

    try:
        deserializer = deserializers[key]
    except KeyError:
        if inspect.isclass(value_type):
            deserializer = class_deserializer
        else:
            raise ValueError(f"Unsupported field of type {value_type}") from None

    # Define the recursion function
    def recur(value: Any, value_type: Type) -> Any:
        return common_deserialize(
            value,
            value_type,
            class_deserializer,
            deserializers,
            user_deserializers,
        )

    # Apply it
    return deserializer(value, value_type, recur)


def all_subclasses(cls: Type, include_cls: bool) -> Iterable[Type]:
    """
    Yields all classes directly on indirectly inheriting from `cls`. Does not
    perform any sort of cycle checks.

    :param cls: The class to start from.
    :param include_cls: Whether to include `cls` itself in the results.
    """

    if include_cls:
        yield cls

    for subclass in cls.__subclasses__():
        yield from all_subclasses(subclass, include_cls=True)


def custom_get_type_hints(typ: Type) -> dict[str, Type]:
    """
    Returns the type hints for the given type, applying some uniserde specific
    logic.
    """
    hints = get_type_hints(typ)

    # Drop any value named '_' if it's just `dataclasses.KW_ONLY`
    if hasattr(typ, "__dataclass_fields__"):
        try:
            val = hints["_"]
        except KeyError:
            pass
        else:
            if val is dataclasses.KW_ONLY:
                del hints["_"]

    return hints



from __future__ import annotations

from typing import *  # type: ignore

import uniserde

from .typedefs import *

__all__ = [
    "Serde",
]


T = TypeVar("T", bound="Serde")


class Serde:
    def as_bson(
        self,
        *,
        as_type: Optional[Type] = None,
        custom_serializers: dict[Type, Callable[[Any], Any]] = {},
    ) -> BsonDoc:
        """
        Serialize the entire instance to BSON, by applying the field serializer
        to each field. Field names are converted to camel case.
        """
        assert as_type is None or issubclass(self.__class__, as_type), as_type
        return uniserde.as_bson(
            self,
            as_type=as_type,
            custom_serializers=custom_serializers,
        )  # type: ignore

    def as_json(
        self,
        *,
        as_type: Optional[Type] = None,
        custom_serializers: dict[Type, Callable[[Any], Any]] = {},
    ) -> JsonDoc:
        """
        Serialize the entire instance to JSON, by applying the field serializer
        to each field. Field names are converted to camel case.
        """
        assert as_type is None or issubclass(self.__class__, as_type), as_type
        return uniserde.as_json(
            self,
            as_type=as_type,
            custom_serializers=custom_serializers,
        )  # type: ignore

    @classmethod
    def from_bson(
        cls: Type[T],
        document: BsonDoc,
        *,
        custom_deserializers: dict[Type, Callable[[Any], Any]] = {},
        lazy: bool = False,
    ) -> T:
        """
        Deserialize an entire data class from BSON, by applying the field
        deserializer to each field. Field names are converted from camel case.

        Warning: The document may be modified in-place by this function!
        """
        return uniserde.from_bson(
            document,
            as_type=cls,
            custom_deserializers=custom_deserializers,
            lazy=lazy,
        )

    @classmethod
    def from_json(
        cls: Type[T],
        document: JsonDoc,
        *,
        custom_deserializers: dict[Type, Callable[[Any], Any]] = {},
        lazy: bool = False,
    ) -> T:
        """
        Deserialize an entire class from JSON, by applying the field
        deserializer to each field. Field names are converted from camel case.

        Warning: The document may be modified in-place by this function!
        """
        return uniserde.from_json(
            document,
            as_type=cls,
            custom_deserializers=custom_deserializers,
            lazy=lazy,
        )

    @classmethod
    def as_mongodb_schema(
        cls,
        *,
        custom_handlers: dict[Type, Callable[[Any], Any]] = {},
    ) -> Any:
        """
        Return a MongoDB schema for this class. The schema matches values
        resulting from `as_bson`.
        """
        return uniserde.as_mongodb_schema(
            cls,
            custom_handlers=custom_handlers,
        )



from __future__ import annotations

from typing import *  # type: ignore
from dataclasses import dataclass
import uniserde
from pathlib import Path


value_raw = Path.home()
value_serialized = uniserde.as_json(value_raw)
print(value_serialized)

value_parsed = uniserde.from_json(value_serialized, Path)
print(value_parsed)


def foo():
    @dataclass
    class Minimum:
        minimum_value: float

    # class Constraint:
    #     def __init__(self, *args, **kwargs) -> None:
    #         pass

    #     @staticmethod
    #     def __call__(*args, **kwargs) -> Any:
    #         pass

    T = TypeVar("T")
    Tup = TypeVarTuple("Tup")

    # Constraint = Annotated[T, Tup]
    Constraint: TypeAlias = Annotated

    MnimumAnn = Annotated[T, Minimum(0.0)]

    class Foo:
        float_value: float

        int_value: Constraint[
            int,
            Minimum(0),
        ]



import enum
import inspect
import types
from datetime import datetime, timedelta
from pathlib import Path
from typing import *  # type: ignore

from . import case_convert, common, serde_class
from .common import *
from .objectid_proxy import ObjectId
from .typedefs import Jsonable


class MongoDbSchemaBuilder:
    def __init__(self, custom_handlers: dict[Type, Callable[[Type], Jsonable]]):
        self._custom_handlers = custom_handlers

    def process(self, value_type: Type) -> Jsonable:
        # Find a matching serializer
        key = common.get_type_key(value_type)

        # Custom handlers take precedence
        try:
            handler = self._custom_handlers[key]
        except KeyError:
            pass
        else:
            return handler(value_type)

        # Plain old default handler
        try:
            handler = self._handlers[key]
        except KeyError:
            pass
        else:
            return handler(self, value_type)

        # Class handler
        assert inspect.isclass(value_type), value_type
        return self._make_schema_class(value_type)

    def _make_schema_bool_to_bool(self, value_type: Type) -> Jsonable:
        return {"type": "boolean"}

    def _make_schema_int_to_int(self, value_type: Type) -> Jsonable:
        return {"bsonType": ["int", "long"]}

    def _make_schema_float_to_float(self, value_type: Type) -> Jsonable:
        return {"bsonType": ["int", "long", "double"]}

    def _make_schema_bytes_to_bytes(self, value_type: Type) -> Jsonable:
        return {"bsonType": "binData"}

    def _make_schema_str_to_str(self, value_type: Type) -> Jsonable:
        return {"type": "string"}

    def _make_schema_datetime_to_datetime(self, value_type: Type) -> Jsonable:
        return {"bsonType": "date"}

    def _make_schema_timedelta_to_float(self, value_type: Type) -> Jsonable:
        return {"bsonType": ["int", "long", "double"]}

    def _make_schema_tuple_to_list(self, value_type: Type) -> Jsonable:
        return {
            "type": "array",
            "items": [self.process(subtype) for subtype in get_args(value_type)],
        }

    def _make_schema_list_to_list(self, value_type: Type) -> Jsonable:
        return {
            "type": "array",
            "items": self.process(get_args(value_type)[0]),
        }

    def _make_schema_set_to_list(self, value_type: Type) -> Jsonable:
        return {
            "type": "array",
            "items": self.process(get_args(value_type)[0]),
        }

    def _make_schema_path_to_str(self, value_type: Type) -> Jsonable:
        return {"type": "string"}

    def _make_schema_dict_to_dict(self, value_type: Type) -> Jsonable:
        subtypes = get_args(value_type)
        assert subtypes[0] is str, value_type

        return {
            "type": "object",
            "items": self.process(subtypes[1]),
        }

    def _make_schema_object_id_to_object_id(self, value_type: Type) -> Jsonable:
        return {"bsonType": "objectId"}

    def _make_schema_literal_to_str(self, value_type: Type) -> Jsonable:
        return {"type": "string"}

    def _make_schema_union(self, value_type: Type) -> Jsonable:
        # Convert each subtype to a BSON schema
        sub_schemas = []
        for subtype in get_args(value_type):
            # Union is used by Python to represent "Optional"
            if subtype is type(None):
                sub_schemas.append({"type": "null"})
                continue

            sub_schemas.append(self.process(subtype))

        # Prettify the result: instead of `{anyof {type ...} {type ...}}` just
        # create one `type`
        types = []
        bson_types = []
        others = []

        for schema in sub_schemas:
            if len(schema) == 1:
                # Standard Json Schema type
                try:
                    type_field = schema["type"]
                except KeyError:
                    pass
                else:
                    if isinstance(type_field, list):
                        types.extend(type_field)
                    else:
                        types.append(type_field)

                    continue

                # BSON type
                try:
                    type_field = schema["bsonType"]
                except KeyError:
                    pass
                else:
                    if isinstance(type_field, list):
                        bson_types.extend(type_field)
                    else:
                        bson_types.append(type_field)

                    continue

            # General case
            others.append(schema)

        # Create new, merged schemas
        sub_schemas = []

        if bson_types:
            sub_schemas.append({"bsonType": types + bson_types})
        elif types:
            sub_schemas.append({"type": types})

        sub_schemas.extend(others)

        if len(sub_schemas) == 1:
            return sub_schemas[0]

        return {"anyOf": sub_schemas}

    def _make_schema_any(self, value_type: Type) -> Jsonable:
        return {}

    def _create_class_schema_ignore_serialize_as_child(
        self, value_type: Type
    ) -> Jsonable:
        doc_field_names = []
        doc_properties = {}
        result = {
            "type": "object",
            "properties": doc_properties,
            "additionalProperties": False,
        }

        for field_py_name, field_type in common.custom_get_type_hints(
            value_type
        ).items():
            field_doc_name = case_convert.all_lower_to_camel_case(field_py_name)

            if field_py_name == "id":
                field_doc_name = "_id"

            doc_field_names.append(field_doc_name)
            doc_properties[field_doc_name] = self.process(field_type)

        # The `required` field may only be present if it contains at least one value
        if doc_field_names:
            result["required"] = doc_field_names

        return result

    def _make_schema_class(self, value_type: Type) -> Jsonable:
        assert inspect.isclass(value_type), value_type

        # Case: The class has a custom schema method
        try:
            override_method = getattr(value_type, "as_mongodb_schema")
        except AttributeError:
            pass
        else:
            serde_class_method = serde_class.Serde.as_mongodb_schema

            try:
                override_method_func = override_method.__func__
            except AttributeError:
                override_method_func = override_method

            if override_method_func is not serde_class_method.__func__:
                return override_method()

        # Case: enum.Flag
        if issubclass(value_type, enum.Flag):
            return {
                "type": "array",
                "items": {
                    "enum": [
                        case_convert.all_upper_to_camel_case(variant.name)  # type: ignore
                        for variant in value_type
                    ],
                },
            }

        # Case: enum.Enum
        if issubclass(value_type, enum.Enum):
            return {
                "enum": [
                    case_convert.all_upper_to_camel_case(variant.name)
                    for variant in value_type
                ],
            }

        # Case: Class, and definitely not one of it's children
        if not should_serialize_as_child(value_type):
            return self._create_class_schema_ignore_serialize_as_child(value_type)

        # Case: Class or one of its children

        # Create the schemas for all allowable classes
        sub_schemas = []
        for subtype in all_subclasses(value_type, True):
            schema: Any = self._create_class_schema_ignore_serialize_as_child(subtype)
            assert schema["type"] == "object", schema

            schema["properties"]["type"] = {
                "enum": [case_convert.upper_camel_case_to_camel_case(subtype.__name__)]
            }

            required = schema.setdefault("required", [])
            required.insert(0, "type")

            sub_schemas.append(schema)

        # Create the final, combined schema
        if len(sub_schemas) == 1:
            return sub_schemas[0]
        else:
            return {"anyOf": sub_schemas}

    _handlers: dict[Type, Callable[["MongoDbSchemaBuilder", Type], Jsonable]] = {
        bool: _make_schema_bool_to_bool,
        int: _make_schema_int_to_int,
        float: _make_schema_float_to_float,
        bytes: _make_schema_bytes_to_bytes,
        str: _make_schema_str_to_str,
        datetime: _make_schema_datetime_to_datetime,
        timedelta: _make_schema_timedelta_to_float,
        list: _make_schema_list_to_list,
        dict: _make_schema_dict_to_dict,
        Union: _make_schema_union,
        Any: _make_schema_any,
        ObjectId: _make_schema_object_id_to_object_id,
        Literal: _make_schema_literal_to_str,
        tuple: _make_schema_tuple_to_list,
        set: _make_schema_set_to_list,
        Path: _make_schema_path_to_str,
        type(Path()): _make_schema_path_to_str,
    }


def as_mongodb_schema(
    value_type: Type,
    *,
    custom_handlers: dict[Type, Callable[[Type], Jsonable]] = {},
) -> Any:
    """
    Return a MongoDB schema for this class. The schema matches values resulting
    from `as_bson`.
    """

    builder = MongoDbSchemaBuilder(custom_handlers)
    return builder.process(value_type)



import re

_SPLIT_CAMEL_CASE_PATTERN = re.compile(
    ".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)


def _is_lower(c: str) -> bool:
    """
    Does the obvious.

    The built-in `str.islower` method is not used because it mysteriously
    returns `False` for some strings, such as `_`.
    """
    return c == c.lower()


def _is_upper(c: str) -> bool:
    """
    Does the obvious.

    The built-in `str.islower` method is not used because it mysteriously
    returns `False` for some strings, such as `_`.
    """
    return c == c.upper()


def all_lower_to_camel_case(name: str) -> str:
    """
    Converts a string from all_lower_case to camelCase.
    """
    assert _is_lower(name), name

    if not name:
        return ""

    parts = name.split("_")
    assert parts, (name, parts)

    return parts[0] + "".join(p.capitalize() for p in parts[1:])


def all_upper_to_camel_case(name: str) -> str:
    """
    Converts a string from ALL_UPPER_CASE to camelCase.
    """
    assert _is_upper(name), name

    if not name:
        return ""

    parts = name.split("_")
    assert parts, (name, parts)

    return parts[0].lower() + "".join(p.capitalize() for p in parts[1:])


def camel_case_to_all_upper(name: str) -> str:
    """
    Converts a string from camelCase to ALL_UPPER_CASE.
    """
    matches = _SPLIT_CAMEL_CASE_PATTERN.finditer(name)
    groups = [m.group(0).upper() for m in matches]
    return "_".join(groups)


def upper_camel_case_to_camel_case(name: str) -> str:
    """
    Converts a string from UpperCamelCase to camelCase.
    """
    if not name:
        return ""

    return name[0].lower() + name[1:]



import base64
import typing
from datetime import datetime, timedelta
from pathlib import Path
from typing import *  # type: ignore

import dateutil.parser

from . import caching_serdeserializer, case_convert, common
from .common import SerdeError
from .objectid_proxy import ObjectId
from .typedefs import Jsonable

__all__ = [
    "from_json",
]


T = TypeVar("T")


class JsonDeserializer(caching_serdeserializer.CachingSerDeserializer[Jsonable, Any]):
    def _get_class_fields(self, value_type: Type) -> Iterable[tuple[str, str, Type]]:
        for field_name_py, field_type in common.custom_get_type_hints(
            value_type
        ).items():
            yield (
                field_name_py,
                case_convert.all_lower_to_camel_case(field_name_py),
                field_type,
            )

    def _deserialize_bytes_from_str(
        self,
        value: Any,
        value_type: Type[str],
    ) -> bytes:
        if not isinstance(value, str):
            raise SerdeError(f"Expected bytes encoded as string, got `{value}`")

        try:
            return base64.b64decode(value.encode("utf-8"))
        except ValueError:
            raise SerdeError("Received invalid base64 encoded string.")

    def _deserialize_datetime_from_str(
        self,
        value: Any,
        value_type: Type[datetime],
    ) -> datetime:
        if not isinstance(value, str):
            raise SerdeError(f"Expected date/time string, got `{value}`")

        try:
            result = dateutil.parser.isoparse(value)
        except ValueError:
            raise SerdeError(f"Expected date/time, got `{value}`") from None

        if result.tzinfo is None:
            raise SerdeError(f"The date/time value `{value}` is missing a timezone.")

        return result

    def _deserialize_timedelta_from_float(
        self,
        value: Any,
        value_type: Type[timedelta],
    ) -> timedelta:
        if not isinstance(value, (int, float)):
            raise SerdeError(f"Expected number, got `{value}`")

        return timedelta(seconds=value)

    def _deserialize_tuple_from_list(
        self,
        value: Any,
        value_type: Type[Tuple],
    ) -> tuple[Any]:
        if not isinstance(value, list):
            raise SerdeError(f"Expected list, got `{value}`")

        subtypes = get_args(value_type)

        if len(value) != len(subtypes):
            raise SerdeError(
                f"Expected list of length {len(subtypes)}, but received one of length {len(value)}"
            )

        return tuple(self.process(v, subtype) for v, subtype in zip(value, subtypes))

    def _deserialize_list_from_list(
        self,
        value: Any,
        value_type: Type[List],
    ) -> list[Any]:
        if not isinstance(value, list):
            raise SerdeError(f"Expected list, got `{value}`")

        subtype = get_args(value_type)[0]
        child_deserializer = self._get_handler(subtype)

        return [child_deserializer(self, v, subtype) for v in value]

    def _deserialize_set_from_list(
        self,
        value: Any,
        value_type: Type[Set],
    ) -> Set:
        if not isinstance(value, list):
            raise SerdeError(f"Expected list, got `{value}`")

        subtype = get_args(value_type)[0]

        return set(self.process(v, subtype) for v in value)

    def _deserialize_dict_from_dict(
        self,
        value: Any,
        value_type: Type[Dict],
    ) -> dict[Any, Any]:
        if not isinstance(value, dict):
            raise SerdeError(f"Expected dict, got `{value}`")

        subtypes = get_args(value_type)

        key_type = subtypes[0]
        key_deserializer = self._get_handler(key_type)

        value_type = subtypes[1]
        value_deserializer = self._get_handler(value_type)

        return {
            key_deserializer(self, k, key_type): value_deserializer(self, v, value_type)
            for k, v in value.items()
        }

    def _deserialize_object_id_from_str(
        self,
        value: Any,
        value_type: Type[ObjectId],
    ) -> ObjectId:
        if not isinstance(value, str):
            raise SerdeError(f"Expected ObjectId string, got `{value}`")

        try:
            result = ObjectId(value)
        except ValueError:
            raise SerdeError(f"Expected ObjectId string, got `{value}`") from None

        return result

    def _deserialize_optional(
        self,
        value: Any,
        value_type: Type,
    ) -> Any:
        if value is None:
            return None

        for subtype in typing.get_args(value_type):
            if subtype is not type(None):
                break
        else:
            assert False, "No `None` in an `Optional` type?"

        return self.process(value, subtype)

    def _deserialize_any(
        self,
        value: Any,
        value_type: Type[Any],
    ) -> Any:
        return value

    def _deserialize_literal_as_is(
        self,
        value: Any,
        value_type: Type[Any],
    ) -> str:
        options = get_args(value_type)
        if value not in options:
            raise SerdeError(f"Expected `{value_type}`, got `{value}`")

        return value

    def _deserialize_path_from_str(
        self,
        value: Any,
        value_type: Type[Path],
    ) -> Path:
        if not isinstance(value, str):
            raise SerdeError(f"Expected path string, got `{value}`")

        return Path(value)

    _passthrough_types = {
        bool,
        int,
        float,
        str,
    }  # type: ignore

    _handler_cache = {
        (True, bytes): _deserialize_bytes_from_str,
        (False, bytes): _deserialize_bytes_from_str,
        (True, datetime): _deserialize_datetime_from_str,
        (False, datetime): _deserialize_datetime_from_str,
        (True, timedelta): _deserialize_timedelta_from_float,
        (False, timedelta): _deserialize_timedelta_from_float,
        (True, tuple): _deserialize_tuple_from_list,
        (False, tuple): _deserialize_tuple_from_list,
        (True, list): _deserialize_list_from_list,
        (False, list): _deserialize_list_from_list,
        (True, set): _deserialize_set_from_list,
        (False, set): _deserialize_set_from_list,
        (True, dict): _deserialize_dict_from_dict,
        (False, dict): _deserialize_dict_from_dict,
        (True, Union): _deserialize_optional,
        (False, Union): _deserialize_optional,
        (True, Any): _deserialize_any,
        (False, Any): _deserialize_any,
        (True, ObjectId): _deserialize_object_id_from_str,
        (False, ObjectId): _deserialize_object_id_from_str,
        (True, Literal): _deserialize_literal_as_is,
        (False, Literal): _deserialize_literal_as_is,
        (True, Path): _deserialize_path_from_str,
        (False, Path): _deserialize_path_from_str,
        (True, type(Path())): _deserialize_path_from_str,
        (False, type(Path())): _deserialize_path_from_str,
    }  # type: ignore

    _override_method_name = "from_json"


def from_json(
    document: Any,
    as_type: Type[T],
    *,
    custom_deserializers: dict[Type, Callable[[Jsonable], Any]] = {},
    lazy: bool = False,
) -> T:
    """
    Deserialize an entire class from JSON, by applying the field deserializer to
    each field. Field names are converted from camel case.

    Warning: The document may be modified in-place by this function!
    """

    deserializer = JsonDeserializer(
        custom_deserializers={
            t: lambda v, _: cb(v) for t, cb in custom_deserializers.items()
        },
        lazy=lazy,
    )

    return deserializer.process(document, as_type)



from datetime import datetime
from typing import *  # type: ignore

from typing_extensions import TypeAlias

from .objectid_proxy import ObjectId

__all__ = [
    "Jsonable",
    "Bsonable",
    "JsonDoc",
    "BsonDoc",
]


Jsonable: TypeAlias = Union[
    None,
    bool,
    int,
    float,
    str,
    dict[str, "Jsonable"],
    list["Jsonable"],
    tuple["Jsonable", ...],
]

Bsonable: TypeAlias = Union[
    None,
    bool,
    int,
    float,
    str,
    dict[str, "Bsonable"],
    list["Bsonable"],
    tuple["Bsonable", ...],
    bytes,
    ObjectId,
    datetime,
]

JsonDoc = dict[str, Jsonable]
BsonDoc = dict[str, Bsonable]



import warnings
from typing import *  # type: ignore

import models
import pytest

import uniserde

# This throws a warning since 3.11, which shows up rather confusingly in pytest.
# Suppress it. (The warning is caused by `mongo_schema`, not uniserde.)
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import mongo_schema

try:
    import bson  # type: ignore
except ImportError:
    bson = None


@pytest.mark.skipif(bson is None, reason="`pymongo` is not installed")
@pytest.mark.parametrize(
    "value,value_type",
    [
        (models.TestClass.create_variant_1(), models.TestClass),
        (models.TestClass.create_variant_2(), models.TestClass),
        (models.ParentClass.create_parent_variant_1(), models.ParentClass),
        (models.ChildClass.create_child_variant_1(), models.ChildClass),
        (models.ChildClass.create_child_variant_1(), models.ParentClass),
        (models.ClassWithId.create(), models.ClassWithId),
        (models.ClassWithKwOnly.create(), models.ClassWithKwOnly),
    ],
)
def test_value_matches_schema(
    value: uniserde.Serde,
    value_type: Type[uniserde.Serde],
) -> None:
    schema = uniserde.as_mongodb_schema(value_type)
    bson_value = value.as_bson()

    mongo_schema.validate(bson_value, schema)


def test_overridden_as_mongodb_schema_staticmethod() -> None:
    value_schema = uniserde.as_mongodb_schema(models.ClassWithStaticmethodOverrides)

    assert value_schema == {
        "value": "overridden value",
        "format": "mongodb schema",
    }


def test_overridden_as_mongodb_schema_classmethod() -> None:
    value_schema = uniserde.as_mongodb_schema(models.ClassWithClassmethodOverrides)

    assert value_schema == {
        "value": "overridden value",
        "format": "mongodb schema",
    }



from datetime import datetime, timezone
from typing import *  # type: ignore

from . import caching_serdeserializer, case_convert, common, json_deserialize
from .common import SerdeError
from .objectid_proxy import ObjectId
from .typedefs import Bsonable

__all__ = [
    "from_bson",
]


T = TypeVar("T")


class BsonDeserializer(caching_serdeserializer.CachingSerDeserializer[Bsonable, Any]):
    def _get_class_fields(self, value_type: Type) -> Iterable[tuple[str, str, Type]]:
        for field_name_py, field_type in common.custom_get_type_hints(
            value_type
        ).items():
            if field_name_py == "id":
                field_name_doc = "_id"
            else:
                field_name_doc = case_convert.all_lower_to_camel_case(field_name_py)

            yield (
                field_name_py,
                field_name_doc,
                field_type,
            )

    def _deserialize_datetime_from_datetime(
        self,
        value: Bsonable,
        value_type: Type[datetime],
    ) -> datetime:
        if not isinstance(value, datetime):
            raise SerdeError(f"Expected datetime, got {value}")

        # BSON doesn't support timezones, and MongoDB convention dictates UTC to
        # be assumed. Impute that.
        if value.tzinfo is None:
            value = value.replace(tzinfo=timezone.utc)

        return value

    _passthrough_types = {
        bool,
        int,
        float,
        bytes,
        str,
        ObjectId,
    }  # type: ignore

    _handler_cache = json_deserialize.JsonDeserializer._handler_cache.copy()  # type: ignore
    _handler_cache.update(
        {
            (True, datetime): _deserialize_datetime_from_datetime,
            (False, datetime): _deserialize_datetime_from_datetime,
        }  # type: ignore
    )

    _override_method_name = "from_bson"


def from_bson(
    document: Any,
    as_type: Type[T],
    *,
    custom_deserializers: dict[Type, Callable[[Bsonable], Any]] = {},
    lazy: bool = False,
) -> T:
    """
    Deserialize an entire data class from BSON, by applying the field
    deserializer to each field. Field names are converted from camel case.

    Warning: The document may be modified in-place by this function!
    """

    deserializer = BsonDeserializer(
        custom_deserializers={
            t: lambda v, _: cb(v) for t, cb in custom_deserializers.items()
        },
        lazy=lazy,
    )

    return deserializer.process(document, as_type)



import base64
import enum
import inspect
import typing
from datetime import datetime, timedelta
from pathlib import Path
from typing import *  # type: ignore

from . import case_convert, common, serde_class
from .common import *
from .objectid_proxy import ObjectId
from .typedefs import JsonDoc

__all__ = [
    "as_json",
]


def serialize_bool_to_bool(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> bool:
    assert isinstance(value, bool), value
    return value


def serialize_int_to_int(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> int:
    assert isinstance(value, int), value
    return value


def serialize_float_to_float(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> float:
    assert isinstance(value, (int, float)), value
    return float(value)


def serialize_bytes_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, bytes), value
    return base64.b64encode(value).decode("utf-8")


def serialize_str_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, str), value
    return value


def serialize_datetime_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, datetime), value
    assert (
        value.tzinfo is not None
    ), f"Encountered datetime without a timezone. Please always set timezones, or expect hard to find bugs."
    return value.isoformat()


def serialize_timedelta_to_float(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> float:
    assert isinstance(value, timedelta), value
    return value.total_seconds()


def serialize_tuple_as_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> list[Any]:
    assert isinstance(value, tuple), value
    subtypes = typing.get_args(value_type)
    assert len(subtypes) == len(value), (subtypes, value)

    return [recur(v, subtype) for v, subtype in zip(value, subtypes)]


def serialize_list_to_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> list[Any]:
    assert isinstance(value, list), value
    return [recur(v, typing.get_args(value_type)[0]) for v in value]


def serialize_set_as_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> list[Any]:
    assert isinstance(value, set), value
    subtype = typing.get_args(value_type)[0]
    return [recur(v, subtype) for v in value]


def serialize_dict_to_dict(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> dict[Any, Any]:
    subtypes = typing.get_args(value_type)
    assert isinstance(value, dict), value
    return {recur(k, subtypes[0]): recur(v, subtypes[1]) for k, v in value.items()}


def serialize_object_id_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, ObjectId), value
    return str(value)


def serialize_optional(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    if value is None:
        return None

    for subtype in typing.get_args(value_type):
        if subtype is not type(None):
            break
    else:
        assert False, "No `None` in an `Optional` type?"

    return recur(value, subtype)


def serialize_any(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    return value


def serialize_literal_as_is(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    assert isinstance(value, str), value
    return value


def serialize_path_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, Path), value
    return str(value.absolute())


def serialize_class(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    assert inspect.isclass(value_type), value_type

    # Case: The class has a custom serialization method
    try:
        override_method = getattr(value, "as_json")
    except AttributeError:
        pass
    else:
        if override_method.__func__ is not serde_class.Serde.as_json:
            return override_method()

    # Case: enum.Flag
    if issubclass(value_type, enum.Flag):
        assert isinstance(value, value_type), value
        return [case_convert.all_upper_to_camel_case(flag.name) for flag in value]  # type: ignore

    # Case: enum.Enum
    if issubclass(value_type, enum.Enum):
        assert isinstance(value, value_type), value
        return case_convert.all_upper_to_camel_case(value.name)

    # Case: Anything else
    # Make sure to serialize as the correct class
    if should_serialize_as_child(value_type):
        assert issubclass(type(value), value_type), (type(value), value_type)
        value_type = type(value)

    result = {}
    for field_py_name, field_type in common.custom_get_type_hints(value_type).items():
        field_doc_name = case_convert.all_lower_to_camel_case(field_py_name)
        result[field_doc_name] = recur(getattr(value, field_py_name), field_type)

    # Add a type tag?
    if should_serialize_as_child(value_type):
        result["type"] = case_convert.upper_camel_case_to_camel_case(
            value.__class__.__name__
        )

    return result


JSON_SERIALIZERS: dict[Type, Serializer] = {
    bool: serialize_bool_to_bool,
    int: serialize_int_to_int,
    float: serialize_float_to_float,
    bytes: serialize_bytes_to_str,
    str: serialize_str_to_str,
    datetime: serialize_datetime_to_str,
    timedelta: serialize_timedelta_to_float,
    tuple: serialize_tuple_as_list,
    list: serialize_list_to_list,
    set: serialize_set_as_list,
    dict: serialize_dict_to_dict,
    Union: serialize_optional,
    Any: serialize_any,
    ObjectId: serialize_object_id_to_str,
    Literal: serialize_literal_as_is,
    Path: serialize_path_to_str,
    type(Path()): serialize_path_to_str,
}


def as_json(
    value: Any,
    *,
    as_type: Optional[Type] = None,
    custom_serializers: dict[Type, Callable[[Any], Any]] = {},
) -> JsonDoc:
    return common_serialize(
        value,
        type(value) if as_type is None else as_type,
        serialize_class,
        JSON_SERIALIZERS,
        custom_serializers,
    )



from .bson_deserialize import from_bson
from .bson_serialize import *
from .common import SerdeError, as_child
from .json_deserialize import from_json
from .json_serialize import *
from .objectid_proxy import ObjectId
from .schema_mongodb import as_mongodb_schema
from .serde_class import Serde
from .typedefs import *



import base64
from datetime import datetime, timezone
from pathlib import Path

import models
import pytest

import uniserde


def test_serialize_exact_variant_1() -> None:
    value_fresh = models.TestClass.create_variant_1()
    value_json = value_fresh.as_json()

    assert value_json == {
        "valBool": value_fresh.val_bool,
        "valInt": value_fresh.val_int,
        "valFloat": value_fresh.val_float,
        "valBytes": base64.b64encode(value_fresh.val_bytes).decode("utf-8"),
        "valStr": value_fresh.val_str,
        "valDatetime": value_fresh.val_datetime.isoformat(),
        "valTimedelta": value_fresh.val_timedelta.total_seconds(),
        "valTuple": list(value_fresh.val_tuple),
        "valList": value_fresh.val_list,
        "valSet": list(value_fresh.val_set),
        "valDict": value_fresh.val_dict,
        "valOptional": value_fresh.val_optional,
        "valOldUnionOptional1": value_fresh.val_old_union_optional_1,
        "valOldUnionOptional2": value_fresh.val_old_union_optional_2,
        "valNewUnionOptional1": value_fresh.val_new_union_optional_1,
        "valNewUnionOptional2": value_fresh.val_new_union_optional_2,
        "valAny": value_fresh.val_any,
        "valObjectId": str(value_fresh.val_object_id),
        "valLiteral": value_fresh.val_literal,
        "valEnum": "one",
        "valFlag": ["one", "two"],
        "valPath": str(value_fresh.val_path),
    }


@pytest.mark.parametrize("lazy", [False, True])
def test_round_trip_variant_1(lazy: bool) -> None:
    value_fresh = models.TestClass.create_variant_1()
    value_json = value_fresh.as_json()
    value_round_trip = models.TestClass.from_json(value_json, lazy=lazy)

    assert value_fresh == value_round_trip


@pytest.mark.parametrize("lazy", [False, True])
def test_round_trip_variant_2(lazy: bool) -> None:
    value_fresh = models.TestClass.create_variant_2()
    value_json = value_fresh.as_json()
    value_round_trip = models.TestClass.from_json(value_json, lazy=lazy)

    assert value_fresh == value_round_trip


@pytest.mark.parametrize("lazy", [False, True])
def test_round_trip_parent(lazy: bool) -> None:
    value_fresh = models.ParentClass.create_parent_variant_1()
    value_json = value_fresh.as_json()
    value_round_trip = models.ParentClass.from_json(value_json, lazy=lazy)

    assert value_fresh == value_round_trip


@pytest.mark.parametrize("lazy", [False, True])
def test_round_trip_child(lazy: bool) -> None:
    value_fresh = models.ChildClass.create_child_variant_1()
    value_json = value_fresh.as_json()
    value_round_trip = models.ChildClass.from_json(value_json, lazy=lazy)

    assert value_fresh == value_round_trip


@pytest.mark.parametrize("lazy", [False, True])
def test_round_trip_child_as_parent(lazy: bool) -> None:
    value_fresh = models.ChildClass.create_child_variant_1()
    value_json = value_fresh.as_json()
    value_round_trip = models.ParentClass.from_json(value_json, lazy=lazy)

    assert isinstance(value_round_trip, models.ChildClass)
    assert value_fresh == value_round_trip


def test_kw_only() -> None:
    value_fresh = models.ClassWithKwOnly(1, bar=2)
    value_bson = value_fresh.as_bson()

    assert "foo" in value_bson
    assert "bar" in value_bson
    assert "_" not in value_bson
    assert len(value_bson) == 2
    assert value_bson["foo"] == 1
    assert value_bson["bar"] == 2


@pytest.mark.parametrize("lazy", [False, True])
def test_datetime_needs_timezone(lazy: bool) -> None:
    with pytest.raises(uniserde.SerdeError, match="is missing a timezone."):
        uniserde.from_json("2020-01-01T01:02:03.000004", datetime, lazy=lazy)


@pytest.mark.parametrize("lazy", [False, True])
def test_datetime_parses_timezone(lazy: bool) -> None:
    value_parsed = uniserde.from_json(
        "2020-01-01T01:02:03.000004Z",
        datetime,
        lazy=lazy,
    )

    assert isinstance(value_parsed, datetime)
    assert value_parsed.tzinfo is not None
    assert value_parsed == datetime(2020, 1, 1, 1, 2, 3, 4, timezone.utc)


@pytest.mark.parametrize("lazy", [False, True])
def test_int_is_float(lazy: bool) -> None:
    uniserde.from_json(1, float, lazy=lazy)


@pytest.mark.parametrize("lazy", [False, True])
def test_paths_are_made_absolute(lazy: bool) -> None:
    path_relative = Path.home() / "folder"
    path_relative = path_relative.relative_to(Path.home())
    assert not path_relative.is_absolute()

    path_absolute = path_relative.absolute()
    assert path_absolute.is_absolute()

    path_serialized = uniserde.as_json(path_relative)
    assert path_serialized == str(path_absolute)

    path_deserialized = uniserde.from_json(path_serialized, Path, lazy=lazy)
    assert path_deserialized == path_absolute


def test_catch_superfluous_value() -> None:
    with pytest.raises(uniserde.SerdeError, match="Superfluous object field"):
        uniserde.from_json(
            {
                "foo": 1,
                "bar": "one",
                "invalidKey": True,
            },
            models.SimpleClass,
        )


def test_overridden_as_json() -> None:
    value_fresh = models.ClassWithStaticmethodOverrides.create()
    value_json = uniserde.as_json(value_fresh)

    assert value_json == {
        "value": "overridden value",
        "format": "json",
    }


@pytest.mark.parametrize("lazy", [False, True])
def test_overridden_from_json_staticmethod(lazy: bool) -> None:
    value_json = {
        "value": "actual value",
        "format": "json",
    }
    value_parsed = uniserde.from_json(
        value_json,
        models.ClassWithStaticmethodOverrides,
        lazy=lazy,
    )

    assert isinstance(value_parsed, models.ClassWithStaticmethodOverrides)
    assert value_parsed.value == "overridden value"
    assert value_parsed.format == "json"


@pytest.mark.parametrize("lazy", [False, True])
def test_overridden_from_json_classmethod(lazy: bool) -> None:
    value_json = {
        "value": "actual value",
        "format": "json",
    }
    value_parsed = uniserde.from_json(
        value_json,
        models.ClassWithClassmethodOverrides,
        lazy=lazy,
    )

    assert isinstance(value_parsed, models.ClassWithClassmethodOverrides)
    assert value_parsed.value == "overridden value"
    assert value_parsed.format == "json"



from datetime import datetime
from typing import *  # type: ignore

from . import json_serialize, serde_class
from .common import *
from .objectid_proxy import ObjectId
from .typedefs import BsonDoc

__all__ = [
    "as_bson",
]


def serialize_bytes_to_bytes(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> bytes:
    assert isinstance(value, bytes), value
    return value


def serialize_datetime_to_datetime(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> datetime:
    assert isinstance(value, datetime), value
    assert (
        value.tzinfo is not None
    ), f"Encountered datetime without a timezone. Please always set timezones, or expect hard to find bugs."

    return value


def serialize_object_id_to_object_id(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> ObjectId:
    assert isinstance(value, ObjectId), value
    return value


def serialize_class(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> BsonDoc:
    result = json_serialize.serialize_class(value, value_type, recur)

    # Case: The class has a custom serialization method
    try:
        override_method = getattr(value, "as_bson")
    except AttributeError:
        pass
    else:
        if override_method.__func__ is not serde_class.Serde.as_bson:
            return override_method()

    # Map "id" to "_id", as is done in MongoDB.
    if isinstance(result, dict) and "_id" not in result:
        try:
            result["_id"] = result.pop("id")
        except KeyError:
            pass

    return result


BSON_SERIALIZERS: dict[Type, Serializer] = json_serialize.JSON_SERIALIZERS.copy()
BSON_SERIALIZERS.update(
    {
        bytes: serialize_bytes_to_bytes,
        datetime: serialize_datetime_to_datetime,
        ObjectId: serialize_object_id_to_object_id,
    }
)


def as_bson(
    value: Any,
    *,
    as_type: Optional[Type] = None,
    custom_serializers: dict[Type, Callable[[Any], Any]] = {},
) -> BsonDoc:
    return common_serialize(
        value,
        type(value) if as_type is None else as_type,
        serialize_class,
        BSON_SERIALIZERS,
        custom_serializers,
    )



import enum
from dataclasses import KW_ONLY, dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import *  # type: ignore

from uniserde import BsonDoc, JsonDoc, ObjectId, Serde, as_child


class RegularEnum(enum.Enum):
    ONE = 1
    TWO = 2
    THREE = 3


class FlagEnum(enum.Flag):
    ONE = 1
    TWO = 2
    FOUR = 4


@dataclass
class SimpleClass(Serde):
    foo: int
    bar: str


@dataclass
class TestClass(Serde):
    val_bool: bool
    val_int: int
    val_float: float
    val_bytes: bytes
    val_str: str
    val_datetime: datetime
    val_timedelta: timedelta
    val_tuple: tuple[int, str]
    val_list: list[int]
    val_set: Set[int]
    val_dict: dict[str, int]
    val_optional: Optional[int]
    val_old_union_optional_1: Union[int, None]
    val_old_union_optional_2: Union[None, int]
    val_new_union_optional_1: int | None
    val_new_union_optional_2: None | int
    val_any: Any
    val_object_id: ObjectId
    val_literal: Literal["one", "two", "three"]
    val_enum: RegularEnum
    val_flag: FlagEnum
    val_path: Path

    @classmethod
    def create_variant_1(cls) -> "TestClass":
        return cls(
            val_bool=True,
            val_int=1,
            val_float=1.0,
            val_bytes=b"these are bytes",
            val_str="this is a string",
            val_datetime=datetime(2020, 1, 2, tzinfo=timezone.utc),
            val_timedelta=timedelta(days=1, seconds=2, microseconds=3),
            val_tuple=(1, "one"),
            val_list=[1, 2, 3],
            val_set={1, 2, 3},
            val_dict={"one": 1, "two": 2},
            val_optional=1,
            val_old_union_optional_1=1,
            val_old_union_optional_2=1,
            val_new_union_optional_1=1,
            val_new_union_optional_2=1,
            val_any="this is an ANY value",
            val_object_id=ObjectId("62bd611fa847c71f1b68fffb"),
            val_literal="one",
            val_enum=RegularEnum.ONE,
            val_flag=FlagEnum.ONE | FlagEnum.TWO,
            val_path=Path.home() / "one",
        )

    @classmethod
    def create_variant_2(cls) -> "TestClass":
        return cls(
            val_bool=False,
            val_int=2,
            val_float=2.0,
            val_bytes=b"these are different bytes",
            val_str="this is another string",
            val_datetime=datetime(2024, 5, 6, tzinfo=timezone.utc),
            val_timedelta=timedelta(days=10, seconds=20, microseconds=30),
            val_tuple=(2, "two"),
            val_list=[4, 5, 6],
            val_set={4, 5, 6},
            val_dict={"three": 3, "four": 4},
            val_optional=None,
            val_old_union_optional_1=None,
            val_old_union_optional_2=None,
            val_new_union_optional_1=None,
            val_new_union_optional_2=None,
            val_any="this is another ANY value",
            val_object_id=ObjectId("62bd6122a847c71f1b68fffc"),
            val_literal="two",
            val_enum=RegularEnum.TWO,
            val_flag=FlagEnum.ONE | FlagEnum.TWO | FlagEnum.FOUR,
            val_path=Path.home() / "two",
        )


@dataclass
@as_child
class ParentClass(Serde):
    parent_int: int
    parent_float: float

    @classmethod
    def create_parent_variant_1(cls) -> "ParentClass":
        return cls(
            parent_int=1,
            parent_float=1.0,
        )


@dataclass
class ChildClass(ParentClass):
    child_float: float
    child_str: str

    @classmethod
    def create_child_variant_1(cls) -> "ChildClass":
        return cls(
            parent_int=1,
            parent_float=1.0,
            child_float=1.0,
            child_str="this is a string",
        )


@dataclass
class ClassWithId(Serde):
    id: int
    foo: int

    @classmethod
    def create(cls) -> "ClassWithId":
        return cls(1, 2)


@dataclass
class ClassWithKwOnly(Serde):
    foo: int

    _: KW_ONLY

    bar: int

    @classmethod
    def create(cls) -> "ClassWithKwOnly":
        return cls(1, bar=2)


@dataclass
class ClassWithStaticmethodOverrides(Serde):
    """
    Class which has uniserde's special methods overridden. This allows to check
    that they are called rather than the default.

    All methods are overridden as @staticmethod.
    """

    value: str
    format: str

    @classmethod
    def create(cls) -> "ClassWithStaticmethodOverrides":
        return cls("stored value", "python")

    def as_json(
        self,
        *,
        as_type: Optional[Type] = None,
        custom_serializers: dict[Type, Callable[[Any], Any]] = {},
    ) -> JsonDoc:
        return {"value": "overridden value", "format": "json"}

    def as_bson(
        self,
        *,
        as_type: Optional[Type] = None,
        custom_serializers: dict[Type, Callable[[Any], Any]] = {},
    ) -> BsonDoc:
        return {"value": "overridden value", "format": "bson"}

    @staticmethod
    def from_json(
        document: dict[str, Any],
        custom_deserializers: dict[Type, Callable[[Any], Any]] = {},
    ) -> "ClassWithStaticmethodOverrides":
        return ClassWithStaticmethodOverrides("overridden value", "json")

    @staticmethod
    def from_bson(
        document: dict[str, Any],
        custom_deserializers: dict[Type, Callable[[Any], Any]] = {},
    ) -> "ClassWithStaticmethodOverrides":
        return ClassWithStaticmethodOverrides("overridden value", "bson")

    @staticmethod
    def as_mongodb_schema(
        custom_handlers: dict[Type, Callable[[Any], Any]] = {},
    ) -> Any:
        return {"value": "overridden value", "format": "mongodb schema"}


@dataclass
class ClassWithClassmethodOverrides(Serde):
    """
    Same as the class above, but with the methods overridden as @classmethod.
    """

    value: str
    format: str

    @classmethod
    def create(cls) -> "ClassWithClassmethodOverrides":
        return cls("stored value", "python")

    @classmethod
    def from_json(
        cls,
        document: dict[str, Any],
        custom_deserializers: dict[Type, Callable[[Any], Any]] = {},
    ) -> "ClassWithClassmethodOverrides":
        return ClassWithClassmethodOverrides("overridden value", "json")

    @classmethod
    def from_bson(
        cls,
        document: dict[str, Any],
        custom_deserializers: dict[Type, Callable[[Any], Any]] = {},
    ) -> "ClassWithClassmethodOverrides":
        return ClassWithClassmethodOverrides("overridden value", "bson")

    @classmethod
    def as_mongodb_schema(
        cls,
        custom_handlers: dict[Type, Callable[[Any], Any]] = {},
    ) -> Any:
        return {"value": "overridden value", "format": "mongodb schema"}
